/********************************** GYRO *********************************/
/* Gyro code provided by Analog Devices, slightly modified for our robot.*/
/*************************************************************************/

void update_angle(){
/* The units of theta are converter LSBs * ms
   1 deg/s = 5mV = 0.256 LSB for ADXRS300
   1 deg = 0.256LSB * 1000ms = 256 (stored in lsb_ms_per_deg)
   To convert theta to degrees, divide by lsb_ms_per_deg. */

  long delta_t_ms, new_time_ms;
  int i;

  time_ms = mseconds();

  for(;;){
    new_time_ms = mseconds();
    delta_t_ms = (new_time_ms - time_ms);

    /* This does the Riemann sum; 
       CCW gyro output polarity is negative */
    theta -= ((float)analog(GYRO_PORT) - offset) * (float)delta_t_ms;

    time_ms = new_time_ms;
    defer();
  }
} /* update_angle() */


void display_angle(){
  for(;;){
    printf("%d degree\n", (int)(CURRENT_ANGLE));
    msleep(200L);
    defer();
  }
} /* display_angle() */


void go_to_angle(float angle){
/* To convert degrees to units of theta, multiply by lsb_ms_per_deg. */
  float angle_target; 
  float error;
  float tolerance;
  int speed;

  tolerance = 1.0 * lsb_ms_per_deg;
  angle_target = angle * lsb_ms_per_deg;
  error = theta - angle_target;

  if ( fabs(error) > tolerance ){
    spin(100 * -sign(error));
    while(fabs(error) > (40.0 * lsb_ms_per_deg)){
      error = theta - angle_target;
    }

    spin(85 * -sign(error));
    while(fabs(error) > (16.0 * lsb_ms_per_deg)){
      error = theta - angle_target;
    }

    spin(70 * -sign(error));
    while(fabs(error) > (2.0 * lsb_ms_per_deg)){
      error = theta - angle_target;
    }

    /* Slam on the brakes */
    spin(100 * sign(error));
    msleep(10L);

    /* Fine adjust either direction as necessary */
    spin(48 * -sign(error));
    while(fabs(error) > tolerance){
       error = theta - angle_target;
       spin(48 * -sign(error));
    }

    spin(0);
    msleep(50L);
  } /* if ( fabs(error)... */
} /* go_to_angle() */


void spin(int speed){
/* CCW if speed > 0, CW if speed < 0, motors off if speed = 0 */
  if (speed > 0) turning_left = 1;
  else turning_right = 1;
  reset_encoder(SHAFT_L);
  reset_encoder(SHAFT_R);
  
  servo(WHEEL, wheel_center - sign((float)speed) * 80);
  motors(LEFT, -speed);
  motors(RIGHT, speed);
  
  if (speed > 0) turning_left = 0;
  else turning_right = 0;
} /* spin() */


void go_straight(float angle, int speed, int distance){
/* Turns to the specified angle, then goes straight for
   the specified distance. (tweak < speed < 100 - tweak) */
  long start_time;
  int tweak = 5;
  float angle_target;
  float error;
  float tolerance;

  tolerance = lsb_ms_per_deg * 0.0008; /* +/-0.5 degree accuracy */
  
  /* first turn the robot */
  go_to_angle(angle);
  
  if (speed > 0) going_forward = 1;
  else going_backward = 1;

  /* then go forward under gyro control for the specified time */
  start_time = mseconds();
  angle_target = angle * lsb_ms_per_deg;
  reset_encoder(SHAFT_R);
  reset_encoder(SHAFT_L);

  while (read_encoder(SHAFT_R) < distance){
    error = theta - angle_target;
    if (theta < (angle_target - tolerance)){
	/* Turn CCW a little */
      servo(WHEEL, wheel_center - sign((float)speed) * 80);
      motors(LEFT, speed - tweak);
	motors(RIGHT, speed + tweak);
    } else if (theta > (angle_target + tolerance)){
      /* Turn CW a little */
      servo(WHEEL, wheel_center + sign((float)speed) * 80);
      motors(LEFT, speed + tweak);
      motors(RIGHT, speed - tweak);
    } else {
      servo(WHEEL, wheel_center);
	motors(LEFT, speed);
      motors(RIGHT, speed);
    }
    /* check a sensor here and terminate the while loop */
  } /* while */

  /* Throw it in reverse briefly to stop quickly */
  servo(WHEEL, wheel_center);
  motors(LEFT, -100);
  motors(RIGHT, -100);
  msleep(20L);
  motors(LEFT, 0);
  motors(RIGHT, 0);
  msleep(20L);
 
  if (speed > 0) going_forward = 0;
  else going_backward = 0;
} /* go_straight() */